/*
Copyright (C) 2001 CyRUS64 (http://www.boob.co.uk)
Copyright (C) 2006 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "RomSettings.h"

#include "Utility/IniFile.h"

#include "Core/ROM.h"
#include "Interface/RomDB.h"

#include "DaedIO.h"

#include <string>
#include <vector>

//*****************************************************************************
//
//*****************************************************************************
class IRomSettingsDB : public CRomSettingsDB
{
	public:
		IRomSettingsDB();
		virtual ~IRomSettingsDB();

		//
		// CRomSettingsDB implementation
		//
		bool			OpenIniFile( const char * filename );
		void			Commit();												// (STRMNNRMN - Write ini back out to disk?)

		bool			GetSettings( const RomID & id, RomSettings * p_settings ) const;
		void			SetSettings( const RomID & id, const RomSettings & settings );

	private:

		//
		// Utility functions
		//
		int				FindEntry( const RomID & id ) const; 		// Find entry in list
		void			AddEntry( const RomID & id, const RomSettings & settings );

		void			OutputSectionDetails(u32 i, FILE * fh);

	private:

		struct section
		{
			RomID				mRomID;
			RomSettings			mSettings;
			bool				bOutput;			// (STRMNNRMN - Has it been written to disk yet?)
		};

		std::vector<section>	mSections;

		bool					mDirty;				// (STRMNNRMN - Changed since read from disk?)
		TCHAR					mFileName[ MAX_PATH + 1 ];
};



//*****************************************************************************
// Singleton creator
//*****************************************************************************
template<> bool	CSingleton< CRomSettingsDB >::Create()
{
	DAEDALUS_ASSERT_Q(mpInstance == NULL);
	
	mpInstance = new IRomSettingsDB();

	return true;
};


//*****************************************************************************
// Constructor
//*****************************************************************************
IRomSettingsDB::IRomSettingsDB()
{
	mSections.clear();
	mDirty = false;
	mFileName[ 0 ] = '\0';
}

//*****************************************************************************
//
//*****************************************************************************
IRomSettingsDB::~IRomSettingsDB()
{
	if ( mDirty )
	{
		Commit();
	}
}

//*****************************************************************************
//	Remove the specified characters from p_string
//*****************************************************************************
static bool	trim( char * p_string, const char * p_trim_chars )
{
	u32 num_trims = strlen( p_trim_chars );
	char * pin = p_string;
	char * pout = p_string;
	bool found;
	while ( *pin )
	{
		char c = *pin;

		found = false;
		for ( u32 i = 0; i < num_trims; i++ )
		{
			if ( p_trim_chars[ i ] == c )
			{
				// Skip
				found = true;
				break;
			}
		}

		if ( found )
		{
			pin++;
		}
		else
		{
			// Copy
			*pout++ = *pin++;
		}
	}
	*pout = '\0';
	return true;
}

//*****************************************************************************
//
//*****************************************************************************
static RomID	RomIDFromString( const char * str )
{
	u32 crc1, crc2, country;
	sscanf( str, "%08x%08x-%02x", &crc1, &crc2, &country );
	return RomID( crc1, crc2, (u8)country );
}

//*****************************************************************************
//
//*****************************************************************************
bool IRomSettingsDB::OpenIniFile( const char * filename )
{
	CIniFile * p_ini_file( CIniFile::Create( filename ) );
	if( p_ini_file == NULL )
	{
		return false;
	}

	//
	// Remember the filename
	//
	strcpy( mFileName, filename );

	for( u32 section_idx = 0; section_idx < p_ini_file->GetNumSections(); ++section_idx )
	{
		const CIniFileSection * p_section( p_ini_file->GetSection( section_idx ) );

		section newsection;

		newsection.mRomID = RomIDFromString( p_section->GetName() );
		//strcpy(newsection.crccheck, readinfo);
		newsection.bOutput = false;
		ROM_ResetSettings( &newsection.mSettings );

		const CIniFileProperty * p_property;
		if( p_section->FindProperty( "Comment", &p_property ) )
		{
			newsection.mSettings.Comment = p_property->GetValue();
		}
		if( p_section->FindProperty( "Info", &p_property ) )
		{
			newsection.mSettings.Info = p_property->GetValue();
		}
		if( p_section->FindProperty( "Name", &p_property ) )
		{
			newsection.mSettings.GameName = p_property->GetValue();
		}
#ifdef DAEDALUS_PSP
		if( p_section->FindProperty( "Preview", &p_property ) )
		{
			newsection.mSettings.Preview = p_property->GetValue();
		}
#endif
		if( p_section->FindProperty( "DisableDynarec", &p_property ) )
		{
			newsection.mSettings.DynarecEnabled = false;
		}
		if( p_section->FindProperty( "DisablePatches", &p_property ) )
		{
			newsection.mSettings.PatchesEnabled = false;
		}
		if( p_section->FindProperty( "CheckTextureHashFrequency", &p_property ) )
		{
			newsection.mSettings.CheckTextureHashFrequency = ROM_GetTextureHashFrequencyFromFrames( atoi( p_property->GetValue() ) );
		}
		if( p_section->FindProperty( "DisableSpeedSync", &p_property ) )
		{
			newsection.mSettings.SpeedSyncEnabled = false;
		}

		if( p_section->FindProperty( "ExpansionPak", &p_property ) )
		{
			newsection.mSettings.ExpansionPakEnabled = true;
		}

		if( p_section->FindProperty( "SaveEep4k", &p_property ) )
		{
			newsection.mSettings.SaveType = SAVE_TYPE_EEP4K;
		}
		if( p_section->FindProperty( "SaveEep16k", &p_property ) )
		{
			newsection.mSettings.SaveType = SAVE_TYPE_EEP16K;
		}
		if( p_section->FindProperty( "SaveSRAM", &p_property ) )
		{
			newsection.mSettings.SaveType = SAVE_TYPE_SRAM;
		}
		if( p_section->FindProperty( "SaveFlashRam", &p_property ) )
		{
			newsection.mSettings.SaveType = SAVE_TYPE_FLASH;
		}

		mSections.push_back(newsection);
	}
	
	delete p_ini_file;
	return true;
}

//*****************************************************************************
//	Write out the .ini file, keeping the original comments intact
//*****************************************************************************
void IRomSettingsDB::Commit()
{
	TCHAR szFileNameOut[MAX_PATH+1];
	TCHAR szFileNameDelete[MAX_PATH+1];
	u32 i;
	FILE * fhIn;
	FILE * fhOut;
	TCHAR szBuf[1024+1];
	char trim_chars[]="{}\n\r"; //remove first and last character

	sprintf(szFileNameOut, "%s.tmp", mFileName);
	sprintf(szFileNameDelete, "%s.del", mFileName);

	fhIn = fopen(mFileName, "r");
	if (fhIn == NULL)
		return;

	fhOut = fopen(szFileNameOut, "w");
	if (fhOut == NULL)
	{
		fclose(fhIn);
		return;
	}

	// Mark all mSections and needing to be written
	for (i = 0; i < mSections.size(); i++)
	{
		mSections[i].bOutput = false;
	}


	while (fgets(szBuf, 1024, fhIn))
	{
		if (szBuf[0] == '{')
		{
			bool found = false;
			// Start of section
			trim(szBuf,trim_chars);

			RomID id = RomIDFromString( szBuf );

			for (i = 0; i < mSections.size(); i++)
			{
				if (mSections[i].bOutput)
					continue;

				if ( mSections[i].mRomID == id )
				{
					// Output this CRC
					OutputSectionDetails(i, fhOut);
					mSections[i].bOutput = true;
					found = true;
					break;
				}
			}
			if (!found)
			{
				// Do what? This should never happen, unless the user
				// replaces the inifile while Daedalus is running!
			}
		}
		else if (szBuf[0] == '/')
		{
			// Comment
			fputs(szBuf, fhOut);
			continue;
		}

	}

	// Input buffer done-  process any new entries!
	for (i = 0; i < mSections.size(); i++)
	{
		// Skip any that have not been done.
		if (!mSections[i].bOutput)
		{
			OutputSectionDetails(i, fhOut);
			mSections[i].bOutput = true;
		}
	}

	fclose(fhOut);
	fclose(fhIn);

	// Create the new file
	IO::File::Move(mFileName, szFileNameDelete);
	IO::File::Move(szFileNameOut, mFileName);
	IO::File::Delete(szFileNameDelete);

	mDirty = false;
}


//*****************************************************************************
//
//*****************************************************************************
void IRomSettingsDB::OutputSectionDetails(u32 i, FILE * fh)
{
	// Generate the CRC-ID for this rom:
	fprintf(fh, "{%08x%08x-%02x}\n", mSections[i].mRomID.CRC[0], mSections[i].mRomID.CRC[1], mSections[i].mRomID.CountryID );
	
	fprintf(fh, "Name=%s\n", mSections[i].mSettings.GameName.c_str());

	if (!mSections[i].mSettings.Comment.empty())		fprintf(fh, "Comment=%s\n", mSections[i].mSettings.Comment.c_str());
	if (!mSections[i].mSettings.Info.empty())			fprintf(fh, "Info=%s\n", mSections[i].mSettings.Info.c_str());
#ifdef DAEDALUS_PSP
	if (!mSections[i].mSettings.Preview.empty())		fprintf(fh, "Preview=%s\n", mSections[i].mSettings.Preview.c_str());
#endif
	if (!mSections[i].mSettings.DynarecEnabled)			fprintf(fh, "DisableDynarec\n");
	if (!mSections[i].mSettings.PatchesEnabled)			fprintf(fh, "DisablePatches\n");
	if (!mSections[i].mSettings.CheckTextureHashFrequency)	fprintf(fh, "CheckTextureHashFrequency=%d\n", ROM_GetTexureHashFrequencyAsFrames(mSections[i].mSettings.CheckTextureHashFrequency));
	if (!mSections[i].mSettings.SpeedSyncEnabled)		fprintf(fh, "DisableSpeedSync\n");
	if (mSections[i].mSettings.ExpansionPakEnabled)		fprintf(fh, "ExpansionPak\n");
	if (mSections[i].mSettings.dwRescanCount != 0)		fprintf(fh, "OSRescanCount=%d\n", mSections[i].mSettings.dwRescanCount);

	switch ( mSections[i].mSettings.SaveType )
	{
	case SAVE_TYPE_UNKNOWN:		break;									// Print nothing
	case SAVE_TYPE_EEP4K:		fprintf(fh, "SaveEep4k\n" ); break;
	case SAVE_TYPE_EEP16K:		fprintf(fh, "SaveEep16k\n" ); break;
	case SAVE_TYPE_SRAM:		fprintf(fh, "SaveSRAM\n" ); break;
	case SAVE_TYPE_FLASH:		fprintf(fh, "SaveFlashRam\n" ); break;
	}

	fprintf(fh, "\n");			// Spacer
}

//*****************************************************************************
// Internal function to add a new entry to the database
//*****************************************************************************
void IRomSettingsDB::AddEntry( const RomID & id, const RomSettings & settings )
{
	// Add new entry!!!
	section newsection;

	newsection.mRomID = id;
	newsection.bOutput = false;
	newsection.mSettings = settings;

	mSections.push_back(newsection);

	mDirty = true;				// Flag to indicate we should be updated
}

//*****************************************************************************
// Internal function to find the entry corresponding to the specified rom. 
//*****************************************************************************
int IRomSettingsDB::FindEntry( const RomID & id ) const
{
	for ( u32 i = 0; i < mSections.size(); i++)
	{
		if ( mSections[i].mRomID == id )
			return i;
	}

	//
	// Not found
	//
	return -1;
}

//*****************************************************************************
// Retreive the settings for the specified rom. Returns false if the rom is 
// not in the database
//*****************************************************************************
bool	IRomSettingsDB::GetSettings( const RomID & id, RomSettings * p_settings ) const
{
	int i = FindEntry( id );
	if ( i >= 0 )
	{
		*p_settings = mSections[ i ].mSettings;
		return true;
	}
	else
	{
		return false;
	}
}

//*****************************************************************************
// Update the settings for the specified rom - creates a new entry if necessary
//*****************************************************************************
void	IRomSettingsDB::SetSettings( const RomID & id, const RomSettings & settings )
{
	int i = FindEntry( id );
	if ( i >= 0 )
	{
		mSections[ i ].mSettings = settings;
	}
	else
	{
		AddEntry( id, settings );
	}
	
	mDirty = true;
}

