/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"

#include "Texture.h"
#include "NativeTexture.h"
#include "DaedMathUtil.h"
#include "PSPColour.h"
#include "PixelFormat.h"
#include "PngUtil.h"

#include "Debug\DBGConsole.h"

#include <pspgu.h>
#include <psputils.h>

using namespace PixelFormats;

//*****************************************************************************
//
//*****************************************************************************
class ITexture : public CTexture
{
	public:
		ITexture();
		virtual ~ITexture();

		bool							Initialise(u32 width, u32 height, ETextureFormat texture_format);
		CNativeTexture *				GetTexture() const					{ return mpTexture; }
		CNativeTexture *				GetMirrorTextureS() const;
		CNativeTexture *				GetMirrorTextureT() const;
		CNativeTexture *				GetMirrorTextureST() const;
		CNativeTexture *				GetRecolouredTexture( c32 colour ) const;

		virtual u32						GetVideoMemoryUsage() const;
		virtual u32						GetSystemMemoryUsage() const;

		// Provides access to "surface"
		bool							StartUpdate(DrawInfo *di);
		void							EndUpdate(DrawInfo *di);

		void							SaveAsPng( const char * filename ) const;

	protected:
		u32								mWidth;			// May differ from mTexture->GetWidth()
		u32								mHeight;		// May differ from mTexture->GetHeight()
		u8 *							mpTexels;
		CNativeTexture *				mpTexture;
		mutable CNativeTexture *		mpMirrorTextureS;
		mutable CNativeTexture *		mpMirrorTextureT;
		mutable CNativeTexture *		mpMirrorTextureST;
		mutable CNativeTexture *		mpRecolouredTexture;
};



//*****************************************************************************
//
//*****************************************************************************
CTexture * CTexture::Create( u32 width, u32 height, ETextureFormat texture_format )
{
	ITexture * pTexture = new ITexture();

	if (!pTexture->Initialise(width, height, texture_format))
	{
		delete pTexture;
		return NULL;
	}

	return pTexture;
}

//*****************************************************************************
//
//*****************************************************************************
ITexture::ITexture()
:	mWidth(0)
,	mHeight(0)
,	mpTexels(NULL)
,	mpTexture(NULL)
,	mpMirrorTextureS(NULL)
,	mpMirrorTextureT(NULL)
,	mpMirrorTextureST(NULL)
,	mpRecolouredTexture(NULL)
{
}

//*****************************************************************************
//
//*****************************************************************************
ITexture::~ITexture(void)
{
	delete [] mpTexels;

	SAFE_RELEASE( mpTexture );
	SAFE_RELEASE( mpMirrorTextureS );
	SAFE_RELEASE( mpMirrorTextureT );
	SAFE_RELEASE( mpMirrorTextureST );
	SAFE_RELEASE( mpRecolouredTexture );
}

//*****************************************************************************
//
//*****************************************************************************
bool ITexture::Initialise( u32 width, u32 height, ETextureFormat texture_format )
{
	DAEDALUS_ASSERT_Q(mpTexture == NULL);

	mWidth = width;
	mHeight = height;
	mpTexture = CNativeTexture::Create( width, height, texture_format );

	u32		buffer_size( mpTexture->GetBytesRequired() );
	mpTexels = new u8[ buffer_size ];

	return mpTexture != NULL;
}

namespace
{
	// Assumes width p_dst = 2*width p_src and height p_dst = 2*height p_src
	template< typename T, bool MirrorS, bool MirrorT >
	void	MirrorTexelsST( T * p_dst, const T * p_src, u32 width, u32 height, u32 src_stride, u32 dst_stride )
	{
		u32	last_pixel( width * 2 - 1 );

		for( u32 y = 0; y < height; ++y )
		{
			// Copy regular pixels
			memcpy( p_dst, p_src, width * sizeof( T ) );

			if( MirrorS )
			{
				for( u32 x = 0; x < width; ++x )
				{
					p_dst[ last_pixel - x ] = p_src[ x ];
				}
			}

			p_dst = reinterpret_cast< T * >( reinterpret_cast< u8 * >( p_dst ) + dst_stride );
			p_src = reinterpret_cast< const T * >( reinterpret_cast< const u8 * >( p_src ) + src_stride );
		}

		if( MirrorT )
		{
			// Copy remaining rows in reverse order
			for( u32 y = 0; y < height; ++y )
			{
				p_src = reinterpret_cast< const T * >( reinterpret_cast< const u8 * >( p_src ) - src_stride );
			
				// Copy regular pixels
				memcpy( p_dst, p_src, width * sizeof( T ) );

				if( MirrorS )
				{
					for( u32 x = 0; x < width; ++x )
					{
						p_dst[ last_pixel - x ] = p_src[ x ];
					}
				}

				p_dst = reinterpret_cast< T * >( reinterpret_cast< u8 * >( p_dst ) + dst_stride );
			}
		}
	}

	template< bool MirrorS, bool MirrorT >
	CNativeTexture *	MirrorTexture( const void * p_src, u32 width, u32 height, u32 src_stride, ETextureFormat texture_format )
	{
		u32		mirror_width( MirrorS ? 2 * width : width );
		u32		mirror_height( MirrorT ? 2 * height : height );

		CNativeTexture * p_mirrored( CNativeTexture::Create( mirror_width, mirror_height, texture_format ) );
		if(p_mirrored != NULL)
		{
			u8 *		p_dst( new u8[ p_mirrored->GetBytesRequired() ] );
			u32			dst_stride( p_mirrored->GetStride() );

			bool	handled( false );

			switch(texture_format)
			{
			case TexFmt_5650:	MirrorTexelsST< Psp::Pf5650, MirrorS, MirrorT >( reinterpret_cast< Psp::Pf5650 * >( p_dst ), reinterpret_cast< const Psp::Pf5650 * >( p_src ), width, height, src_stride, dst_stride ); handled = true; break;
			case TexFmt_5551:	MirrorTexelsST< Psp::Pf5551, MirrorS, MirrorT >( reinterpret_cast< Psp::Pf5551 * >( p_dst ), reinterpret_cast< const Psp::Pf5551 * >( p_src ), width, height, src_stride, dst_stride ); handled = true; break;
			case TexFmt_4444:	MirrorTexelsST< Psp::Pf4444, MirrorS, MirrorT >( reinterpret_cast< Psp::Pf4444 * >( p_dst ), reinterpret_cast< const Psp::Pf4444 * >( p_src ), width, height, src_stride, dst_stride ); handled = true; break;
			case TexFmt_8888:	MirrorTexelsST< Psp::Pf8888, MirrorS, MirrorT >( reinterpret_cast< Psp::Pf8888 * >( p_dst ), reinterpret_cast< const Psp::Pf8888 * >( p_src ), width, height, src_stride, dst_stride ); handled = true; break;
			}

			DAEDALUS_ASSERT( handled, "Unhandled format" );

			p_mirrored->SetData( p_dst, p_mirrored->GetWidth(), p_mirrored->GetHeight() );

			delete [] p_dst;
		}

		return p_mirrored;
	}
}

//*****************************************************************************
//
//*****************************************************************************
CNativeTexture *	ITexture::GetMirrorTextureS() const
{
	if(mpMirrorTextureS == NULL)
	{
		mpMirrorTextureS = MirrorTexture< true, false >( mpTexels, mWidth, mHeight, mpTexture->GetStride(), mpTexture->GetFormat() );
	}

	return mpMirrorTextureS;
}

//*****************************************************************************
//
//*****************************************************************************
CNativeTexture *	ITexture::GetMirrorTextureT() const
{
	if(mpMirrorTextureT == NULL)
	{
		mpMirrorTextureT = MirrorTexture< false, true >( mpTexels, mWidth, mHeight, mpTexture->GetStride(), mpTexture->GetFormat() );
	}

	return mpMirrorTextureT;
}


//*****************************************************************************
//
//*****************************************************************************
CNativeTexture *	ITexture::GetMirrorTextureST() const
{
	if(mpMirrorTextureST == NULL)
	{
		mpMirrorTextureST = MirrorTexture< true, true >( mpTexels, mWidth, mHeight, mpTexture->GetStride(), mpTexture->GetFormat() );
	}

	return mpMirrorTextureST;
}


//*****************************************************************************
//
//*****************************************************************************
CNativeTexture *	ITexture::GetRecolouredTexture( c32 colour ) const
{
	// XXXX this ignores the colour argument if different to the last call
	if(mpRecolouredTexture == NULL)
	{
		DAEDALUS_ASSERT( mpTexture != NULL, "There is no existing texture" );
		mpRecolouredTexture = mpTexture->Recolour( colour );
	}

	return mpRecolouredTexture;
}

//*****************************************************************************
// Lock the texture so we can start updating it
//*****************************************************************************
bool ITexture::StartUpdate(DrawInfo *di)
{
	if (mpTexture != NULL && mpTexels != NULL)
	{
		// Return a temporary buffer to use
		di->pSurface = mpTexels;
		di->Width = mWidth;
		di->Height = mHeight;
		di->Pitch = mpTexture->GetStride();
		return true;
	}

	return false;
}

//*****************************************************************************
// Finish updating the texture and copy the memory to the native surface
//*****************************************************************************
void ITexture::EndUpdate(DrawInfo *di)
{
	if (mpTexture != NULL)
	{
		mpTexture->SetData( di->pSurface, mpTexture->GetCorrectedWidth(), mpTexture->GetCorrectedHeight() );
	}
}

//*****************************************************************************
//
//*****************************************************************************
u32	ITexture::GetVideoMemoryUsage() const
{
	u32		usage( 0 );

	if(mpTexture != NULL)
	{
		usage += mpTexture->GetVideoMemoryUsage();
	}

	if(mpRecolouredTexture != NULL)
	{
		usage += mpRecolouredTexture->GetVideoMemoryUsage();
	}

	return 0;
}

//*****************************************************************************
//
//*****************************************************************************
u32	ITexture::GetSystemMemoryUsage() const
{
	u32	usage( 0 );

	if(mpTexture != NULL)
	{
		usage += mpTexture->GetBytesRequired();		// mpTexels
		usage += mpTexture->GetSystemMemoryUsage();
	}

	if(mpRecolouredTexture != NULL)
	{
		usage += mpRecolouredTexture->GetSystemMemoryUsage();
	}

	return usage;
}

//*****************************************************************************
//
//*****************************************************************************
void	ITexture::SaveAsPng( const char * filename ) const
{
	PngSaveImage( filename, mpTexels, mpTexture->GetFormat(), mpTexture->GetStride(), mWidth, mHeight );
}
