/*
Copyright (C) 2006 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "GlobalSettingsComponent.h"

#include "UIContext.h"
#include "UIScreen.h"
#include "UISetting.h"
#include "AdjustDeadzoneScreen.h"
#include "PSPGraphics/DrawText.h"
#include "PSPGraphics/PSPColour.h"
#include "PSPGraphics/DaedalusGraphics.h"

#include "Utility/FramerateLimiter.h"

#include "Input/InputManager.h"

#include <pspctrl.h>
#include <pspgu.h>

namespace
{
	const u32		TEXT_AREA_TOP = 15+16+16;
	const u32		TEXT_AREA_LEFT = 40;
	const u32		TEXT_AREA_RIGHT = 440;

	const s32		DESCRIPTION_AREA_TOP = 0;		// We render text aligned from the bottom, so this is largely irrelevant
	const s32		DESCRIPTION_AREA_BOTTOM = 272-10;
	const s32		DESCRIPTION_AREA_LEFT = 16;
	const s32		DESCRIPTION_AREA_RIGHT = 480-16;

	class CViewPortSetting : public CUISetting
	{
	public:
		CViewPortSetting( const char * name, const char * description )
			:	CUISetting( name, description )
		{
		}

		virtual	void	OnNext()				{ gViewportType = EViewportType( (gViewportType+1) % NUM_VIEWPORT_TYPES ); }
		virtual	void	OnPrevious()			{ gViewportType = EViewportType( (gViewportType + NUM_VIEWPORT_TYPES - 1) % NUM_VIEWPORT_TYPES ); }

		virtual const char *	GetSettingName() const
		{
			switch( gViewportType )
			{
			case VT_UNSCALED_4_3:	return "4:3 unscaled (320x240)";
			case VT_SCALED_4_3:		return "4:3 scaled (362x272)";
			case VT_FULLSCREEN:		return "Fullscreen (480x272)";

			case NUM_VIEWPORT_TYPES:
				break;
			}
			DAEDALUS_ERROR( "Unhandled viewport type" );
			return "?";
		}

	};

	class CLimitFramerateSetting : public CUISetting
	{
	public:
		CLimitFramerateSetting( const char * name, const char * description )
			:	CUISetting( name, description )
		{
		}

		virtual	void			OnNext()				{ FramerateLimiter_SetLimit( !FramerateLimiter_GetLimit() ); }
		virtual	void			OnPrevious()			{ FramerateLimiter_SetLimit( !FramerateLimiter_GetLimit() ); }

		virtual const char *	GetSettingName() const	{ return FramerateLimiter_GetLimit() ? "Yes" : "No"; }
	};

	class CAdjustDeadzoneSetting : public CUISetting
	{
	public:
		CAdjustDeadzoneSetting( CUIContext * p_context, const char * name, const char * description )
			:	CUISetting( name, description )
			,	mpContext( p_context )
		{
		}

		virtual	void			OnSelected()
		{
			CAdjustDeadzoneScreen *	adjust_deadzone( CAdjustDeadzoneScreen::Create( mpContext ) );
			adjust_deadzone->Run();
			delete adjust_deadzone;
		}

		virtual const char *	GetSettingName() const
		{
			f32 min_deadzone, max_deadzone;
			CInputManager::Get()->GetStickDeadzones( &min_deadzone, &max_deadzone );

			return DSPrintf( "%d/%d", s32( 100.0f * min_deadzone ), s32( 100.0f * max_deadzone ) );
		}

	private:
		CUIContext *			mpContext;
	};


	class CAdjustFrameskipSetting : public CUISetting
	{
	public:
		CAdjustFrameskipSetting( const char * name, const char * description )
			:	CUISetting( name, description )
		{
		}

		virtual	void			OnNext()				{ gFrameskipValue = EFrameskipValue( (gFrameskipValue + 1) % NUM_FRAMESKIP_VALUES ); }
		virtual	void			OnPrevious()			{ gFrameskipValue = EFrameskipValue( (gFrameskipValue + NUM_FRAMESKIP_VALUES - 1) % NUM_FRAMESKIP_VALUES ); }

		virtual const char *	GetSettingName() const
		{
			switch( gFrameskipValue )
			{
			case FV_DISABLED:		return "Disabled";
			case FV_1:				return "1";
			case FV_2:				return "2";
			case FV_3:				return "3";
			case FV_4:				return "4";
			case FV_5:				return "5";
			case FV_6:				return "6";
			case FV_7:				return "7";
			case FV_8:				return "8";
			case FV_9:				return "9";
			case FV_10:				return "10";

			case NUM_FRAMESKIP_VALUES:
				break;
			}
			DAEDALUS_ERROR( "Unhandled frameskip value" );
			return "?";
		}
	};

	class CAdjustFrequencySetting : public CUISetting
	{
	public:
		CAdjustFrequencySetting( bool * setting, const char * name, const char * description )
			:	CUISetting( name, description )
			,	mSetting( setting )
		{
		}

		virtual bool			IsReadOnly() const		{ return !gAudioPluginEnabled; }		// Disable this if no audio enabled;

		virtual	void			OnNext()				{ if( !IsReadOnly() ) *mSetting = !*mSetting; }
		virtual	void			OnPrevious()			{ if( !IsReadOnly() ) *mSetting = !*mSetting; }

		virtual const char *	GetSettingName() const	{ return (*mSetting) ? "Enabled" : "Disabled"; }

	private:
		bool *					mSetting;
	};

}

extern bool		gDisplayFramerate;				// XXXX
extern bool		gSoftwareClipping;
extern bool		gUseVFPUTnL;
extern bool		gHighlightInexactBlendModes;

//*************************************************************************************
//
//*************************************************************************************
class IGlobalSettingsComponent : public CGlobalSettingsComponent
{
	public:

		IGlobalSettingsComponent( CUIContext * p_context );
		~IGlobalSettingsComponent();

		// CUIComponent
		virtual void				Update( float elapsed_time, const v2 & stick, u32 old_buttons, u32 new_buttons );
		virtual void				Render();

	private:
		CUIElementBag				mElements;
		u32							mSelectedIdx;
};

//*************************************************************************************
//
//*************************************************************************************
CGlobalSettingsComponent::CGlobalSettingsComponent( CUIContext * p_context )
:	CUIComponent( p_context )
{
}

//*************************************************************************************
//
//*************************************************************************************
CGlobalSettingsComponent::~CGlobalSettingsComponent()
{
}

//*************************************************************************************
//
//*************************************************************************************
CGlobalSettingsComponent *	CGlobalSettingsComponent::Create( CUIContext * p_context )
{
	return new IGlobalSettingsComponent( p_context );
}

//*************************************************************************************
//
//*************************************************************************************
IGlobalSettingsComponent::IGlobalSettingsComponent( CUIContext * p_context )
:	CGlobalSettingsComponent( p_context )
,	mSelectedIdx( 0 )
{
	mElements.Add( new CBoolSetting( &gDisplayFramerate, "Display Framerate", "Whether to show the framerate while the rom is running.", "Yes", "No" ) );
	mElements.Add( new CViewPortSetting( "Viewport Size", "The size of the viewport on the PSP." ) );
	mElements.Add( new CLimitFramerateSetting( "Limit Framerate", "This determines whether the refresh rate is limited to 50Hz (for PAL games) or 60Hz (for NTSC games)." ) );
	mElements.Add( new CAdjustFrameskipSetting( "Frameskip", "This determines how many frames are skipped before rendering a new frame. Increasing this value should give a small speedup, at the expense of more jerky graphics." ) );
	mElements.Add( new CBoolSetting( &gAudioPluginEnabled, "Audio", "Whether or not to enable audio emulation.", "Enabled", "Disabled" ) );
	mElements.Add( new CAdjustFrequencySetting( &gAdaptFrequency, " Adjust Frequency", "When enabled, this mode tries to avoid gaps in the audio by adjusting the pitch of the audio stream." ) );
	mElements.Add( new CAdjustDeadzoneSetting( mpContext, "Stick Deadzone", "Adjust the size of the deadzone applied to the PSP stick while playing. Press Start/X to edit." ) );

	mElements.Add( new CBoolSetting( &gSoftwareClipping, "Software Clipping",	"Enable software clipping of vertices. Disable this for a small speedup.", "Enabled", "Disabled" ) );

#ifndef DAEDALUS_PUBLIC_RELEASE
	mElements.Add( new CBoolSetting( &gHighlightInexactBlendModes, "Highlight Inexact Blend Modes",	"Replace inexact blend modes with a placeholder texture.", "Yes", "No" ) );
	mElements.Add( new CBoolSetting( &gUseVFPUTnL, "VFPU TnL",	"Enable VFPU acceleration of vertex transform and lighting.", "Enabled", "Disabled" ) );
#endif
}

//*************************************************************************************
//
//*************************************************************************************
IGlobalSettingsComponent::~IGlobalSettingsComponent()
{
}

//*************************************************************************************
//
//*************************************************************************************
void	IGlobalSettingsComponent::Update( float elapsed_time, const v2 & stick, u32 old_buttons, u32 new_buttons )
{
	if(old_buttons != new_buttons)
	{
		if( new_buttons & PSP_CTRL_UP )
		{
			if( mSelectedIdx > 0 )
			{
				mSelectedIdx = mSelectedIdx - 1;
			}
		}
		if( new_buttons & PSP_CTRL_DOWN )
		{
			if( mSelectedIdx < mElements.GetNumElements()-1 )
			{
				mSelectedIdx = mSelectedIdx + 1;
			}
		}

		if( mSelectedIdx < mElements.GetNumElements() )
		{
			CUIElement *	element( mElements.GetElement( mSelectedIdx ) );

			if( new_buttons & PSP_CTRL_LEFT )
			{
				element->OnPrevious();
			}
			if( new_buttons & PSP_CTRL_RIGHT )
			{
				element->OnNext();
			}
			if( new_buttons & (PSP_CTRL_CROSS|PSP_CTRL_START) )
			{
				element->OnSelected();
			}
		}
	}
}

//*************************************************************************************
//
//*************************************************************************************
void	IGlobalSettingsComponent::Render()
{
	mElements.Draw( mpContext, TEXT_AREA_LEFT, TEXT_AREA_RIGHT, TEXT_AREA_TOP, mSelectedIdx );

	if( mSelectedIdx < mElements.GetNumElements() )
	{
		const char *		p_description( mElements.GetElement( mSelectedIdx )->GetDescription() );
		mpContext->DrawTextArea( DESCRIPTION_AREA_LEFT,
								 DESCRIPTION_AREA_TOP,
								 DESCRIPTION_AREA_RIGHT - DESCRIPTION_AREA_LEFT,
								 DESCRIPTION_AREA_BOTTOM - DESCRIPTION_AREA_TOP,
								 p_description,
								 DrawTextUtilities::TextWhite,
								 CUIContext::VA_BOTTOM );
	}
}

