/*
Copyright (C) 2007 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "PauseScreen.h"

#include "UIContext.h"
#include "UIScreen.h"
#include "AboutComponent.h"
#include "GlobalSettingsComponent.h"
#include "RomSettingsComponent.h"
#include "PauseOptionsComponent.h"

#include "PSPGraphics/FFXMaths.h"
#include "PSPGraphics/DrawText.h"
#include "PSPGraphics/PSPColour.h"
#include "PSPGraphics/GraphicsContext.h"

#include "Core/ROM.h"
#include "Core/CPU.h"
#include "Core/RomSettings.h"

#include "DaedMathUtil.h"

#include <pspctrl.h>
#include <pspgu.h>

#include <string>

namespace 
{
	const u32				TEXT_AREA_TOP = 10;
	const u32				TEXT_AREA_LEFT = 20;
	const u32				TEXT_AREA_RIGHT = 480-20;

	enum EMenuOption
	{
		MO_GLOBAL_SETTINGS = 0,
		MO_ROM_SETTINGS,
		MO_PAUSE_OPTIONS,
		MO_ABOUT,
		NUM_MENU_OPTIONS,
	};

	const EMenuOption	MO_FIRST_OPTION = MO_GLOBAL_SETTINGS;
	const EMenuOption	MO_LAST_OPTION = MO_ABOUT;

	const char * const	gMenuOptionNames[ NUM_MENU_OPTIONS ] =
	{
		"Global Settings",
		"Rom Settings",
		"Paused",
		"About",
	};
}

//*************************************************************************************
//
//*************************************************************************************
class IPauseScreen : public CPauseScreen, public CUIScreen
{
	public:

		IPauseScreen( CUIContext * p_context );
		~IPauseScreen();

		// CPauseScreen
		virtual void				Run();

		// CUIScreen
		virtual void				Update( float elapsed_time, const v2 & stick, u32 old_buttons, u32 new_buttons );
		virtual void				Render();
		virtual bool				IsFinished() const									{ return mIsFinished; }


	private:
		static	EMenuOption			GetNextOption( EMenuOption option )					{ return EMenuOption( (option + 1) % NUM_MENU_OPTIONS ); }
		static	EMenuOption			GetPreviousOption( EMenuOption option )				{ return EMenuOption( (option + NUM_MENU_OPTIONS -1) % NUM_MENU_OPTIONS ); }

				EMenuOption			GetPreviousOption() const							{ return GetPreviousOption( mCurrentOption ); }
				EMenuOption			GetCurrentOption() const							{ return mCurrentOption; }
				EMenuOption			GetNextOption() const								{ return GetNextOption( mCurrentOption ); }

				EMenuOption			GetPreviousValidOption() const;
				EMenuOption			GetNextValidOption() const;

				bool				IsOptionValid( EMenuOption option ) const;

				void				OnSettingsConfirmed( const RomSettings & settings );
				void				OnResume();
				void				OnReset();

	private:
		bool						mIsFinished;

		EMenuOption					mCurrentOption;

		CUIComponent *				mOptionScreens[ NUM_MENU_OPTIONS ];
		CRomSettingsComponent *		mRomSettingsComponent;
};

//*************************************************************************************
//
//*************************************************************************************
CPauseScreen::~CPauseScreen()
{
}

//*************************************************************************************
//
//*************************************************************************************
CPauseScreen *	CPauseScreen::Create( CUIContext * p_context )
{
	return new IPauseScreen( p_context );
}

//*************************************************************************************
//
//*************************************************************************************
IPauseScreen::IPauseScreen( CUIContext * p_context )
:	CUIScreen( p_context )
,	mIsFinished( false )
,	mCurrentOption( MO_PAUSE_OPTIONS )
{
	for( u32 i = 0; i < NUM_MENU_OPTIONS; ++i )
	{
		mOptionScreens[ i ] = NULL;
	}

	mRomSettingsComponent = CRomSettingsComponent::Create( mpContext, new CMemberFunctor1< IPauseScreen, const RomSettings & >( this, &IPauseScreen::OnSettingsConfirmed ) );

	mOptionScreens[ MO_GLOBAL_SETTINGS ]	= CGlobalSettingsComponent::Create( mpContext );
	mOptionScreens[ MO_ROM_SETTINGS ]		= mRomSettingsComponent;
	mOptionScreens[ MO_PAUSE_OPTIONS ]		= CPauseOptionsComponent::Create( mpContext, new CMemberFunctor< IPauseScreen >( this, &IPauseScreen::OnResume ), new CMemberFunctor< IPauseScreen >( this, &IPauseScreen::OnReset ) );
	mOptionScreens[ MO_ABOUT ]				= CAboutComponent::Create( mpContext );

	for( u32 i = 0; i < NUM_MENU_OPTIONS; ++i )
	{
		DAEDALUS_ASSERT( mOptionScreens[ i ] != NULL, "Unhandled screen" );
	}
}

//*************************************************************************************
//
//*************************************************************************************
IPauseScreen::~IPauseScreen()
{
	for( u32 i = 0; i < NUM_MENU_OPTIONS; ++i )
	{
		delete mOptionScreens[ i ];
	}
}


//*************************************************************************************
//
//*************************************************************************************
EMenuOption		IPauseScreen::GetPreviousValidOption() const
{
	bool			looped( false );
	EMenuOption		current_option( mCurrentOption );

	do
	{
		current_option = GetPreviousOption( current_option );
		looped = current_option == mCurrentOption;
	}
	while( !IsOptionValid( current_option ) && !looped );

	return current_option;
}

//*************************************************************************************
//
//*************************************************************************************
EMenuOption		IPauseScreen::GetNextValidOption() const
{
	bool			looped( false );
	EMenuOption		current_option( mCurrentOption );

	do
	{
		current_option = GetNextOption( current_option );
		looped = current_option == mCurrentOption;
	}
	while( !IsOptionValid( current_option ) && !looped );

	return current_option;
}

//*************************************************************************************
//
//*************************************************************************************
bool	IPauseScreen::IsOptionValid( EMenuOption option ) const
{
	return true;
}


//*************************************************************************************
//
//*************************************************************************************
void	IPauseScreen::Update( float elapsed_time, const v2 & stick, u32 old_buttons, u32 new_buttons )
{
	if(old_buttons != new_buttons)
	{
		if(new_buttons & PSP_CTRL_LTRIGGER)
		{
			mCurrentOption = GetPreviousValidOption();
			new_buttons &= ~PSP_CTRL_LTRIGGER;
		}
		if(new_buttons & PSP_CTRL_RTRIGGER)
		{
			mCurrentOption = GetNextValidOption();
			new_buttons &= ~PSP_CTRL_RTRIGGER;
		}

		if(new_buttons & PSP_CTRL_SELECT)
		{
			mIsFinished = true;
			new_buttons &= ~PSP_CTRL_SELECT;
		}
	}

	mOptionScreens[ mCurrentOption ]->Update( elapsed_time, stick, old_buttons, new_buttons );
}

//*************************************************************************************
//
//*************************************************************************************
void	IPauseScreen::Render()
{
	mpContext->ClearBackground( c32( 94, 188, 94 ) );

	s32			x,y;

	y = TEXT_AREA_TOP;

	const char * p_option_text;

	c32		valid_colour( mpContext->GetDefaultTextColour() );
	c32		invalid_colour( 200, 200, 200 );

	EMenuOption		previous( GetPreviousOption() );
	EMenuOption		current( GetCurrentOption() );
	EMenuOption		next( GetNextOption() );
		
	p_option_text = gMenuOptionNames[ previous ];
	x = DrawTextUtilities::AlignText( TEXT_AREA_LEFT, TEXT_AREA_RIGHT, p_option_text, DrawTextUtilities::AT_LEFT );
	mpContext->DrawText( x, y, p_option_text, IsOptionValid( previous ) ? valid_colour : invalid_colour );

	p_option_text = gMenuOptionNames[ current ];
	x = DrawTextUtilities::AlignText( TEXT_AREA_LEFT, TEXT_AREA_RIGHT, p_option_text, DrawTextUtilities::AT_CENTRE );
	mpContext->DrawText( x, y, p_option_text, IsOptionValid( current ) ? valid_colour : invalid_colour );

	p_option_text = gMenuOptionNames[ next ];
	x = DrawTextUtilities::AlignText( TEXT_AREA_LEFT, TEXT_AREA_RIGHT, p_option_text, DrawTextUtilities::AT_RIGHT );
	mpContext->DrawText( x, y, p_option_text, IsOptionValid( next ) ? valid_colour : invalid_colour );

	mOptionScreens[ mCurrentOption ]->Render();
}

//*************************************************************************************
//
//*************************************************************************************
void	IPauseScreen::Run()
{
	mIsFinished = false;
	CUIScreen::Run();

	// Clear everything to black - looks a bit tidier
	CGraphicsContext::Get()->ClearAllSurfaces();
}

//*************************************************************************************
//
//*************************************************************************************
void	IPauseScreen::OnSettingsConfirmed( const RomSettings & settings )
{
	mIsFinished = true;
}

//*************************************************************************************
//
//*************************************************************************************
void IPauseScreen::OnResume()
{
	mIsFinished = true;
}

//*************************************************************************************
//
//*************************************************************************************
void IPauseScreen::OnReset()
{
	CPU_Halt("Resetting");
	mIsFinished = true;
}
