/*
Copyright (C) 2006 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "RomSettingsComponent.h"

#include "UIContext.h"
#include "UIScreen.h"
#include "UISetting.h"
#include "PSPGraphics/DrawText.h"
#include "PSPGraphics/PSPColour.h"

#include "Core/ROM.h"

#include "Input/InputManager.h"

#include <pspctrl.h>
#include <pspgu.h>
namespace
{
	const u32		TEXT_AREA_TOP = 15+16+16;
	const u32		TEXT_AREA_LEFT = 40;
	const u32		TEXT_AREA_RIGHT = 440;

	const s32		DESCRIPTION_AREA_TOP = 0;		// We render text aligned from the bottom, so this is largely irrelevant
	const s32		DESCRIPTION_AREA_BOTTOM = 272-10;
	const s32		DESCRIPTION_AREA_LEFT = 16;
	const s32		DESCRIPTION_AREA_RIGHT = 480-16;

	const u32		MAX_TEXTURE_HASH_FREQUENCY = 10;


	class CSaveTypeSetting : public CUISetting
	{
	public:
		CSaveTypeSetting( ESaveType * setting, const char * name, const char * description )
			:	CUISetting( name, description )
			,	mSetting( setting )
		{
		}

		virtual bool			IsReadOnly() const		{ return true; }

		virtual	void			OnNext()				{ if( !IsReadOnly() ) *mSetting = ESaveType( ( *mSetting + 1 ) % NUM_SAVE_TYPES ); }
		virtual	void			OnPrevious()			{ if( !IsReadOnly() ) *mSetting = ESaveType( ( *mSetting + NUM_SAVE_TYPES - 1 ) % NUM_SAVE_TYPES ); }

		virtual const char *	GetSettingName() const	{ return *mSetting == SAVE_TYPE_UNKNOWN ? "None" : ROM_GetSaveTypeName( *mSetting ); }

	private:
		ESaveType *				mSetting;
	};


	class CTextureHashFrequency : public CUISetting
	{
	public:
		CTextureHashFrequency( ETextureHashFrequency * setting, const char * name, const char * description )
			:	CUISetting( name, description )
			,	mSetting( setting )
		{
		}

		virtual	void			OnNext()				{ *mSetting = ETextureHashFrequency( ( *mSetting + 1 ) % NUM_THF ); }
		virtual	void			OnPrevious()			{ *mSetting = ETextureHashFrequency( ( *mSetting + NUM_THF - 1 ) % NUM_THF ); }

		virtual const char *	GetSettingName() const	{ return ROM_GetTextureHashFrequencyDescription( *mSetting ); }

	private:
		ETextureHashFrequency *	mSetting;
	};



	class CAdjustControllerSetting : public CUISetting
	{
	public:
		CAdjustControllerSetting( u32 * setting, const char * name )
			:	CUISetting( name, "" )
			,	mSetting( setting )
		{
		}

		virtual	void			OnNext()				{ *mSetting = (*mSetting + 1) % CInputManager::Get()->GetNumConfigurations(); }
		virtual	void			OnPrevious()			{ *mSetting = (*mSetting + CInputManager::Get()->GetNumConfigurations() - 1) % CInputManager::Get()->GetNumConfigurations(); }

		virtual const char *	GetSettingName() const	{ return CInputManager::Get()->GetConfigurationName( *mSetting ); }
		virtual const char *	GetDescription() const	{ return CInputManager::Get()->GetConfigurationDescription( *mSetting ); }

	private:
		u32 *					mSetting;
	};

	class CAdjustFrequencySetting : public CUISetting
	{
	public:
		CAdjustFrequencySetting( bool * setting, const char * name, const char * description )
			:	CUISetting( name, description )
			,	mSetting( setting )
		{
		}

		virtual bool			IsReadOnly() const		{ return !gAudioPluginEnabled; }		// Disable this if no audio enabled;

		virtual	void			OnNext()				{ if( !IsReadOnly() ) *mSetting = !*mSetting; }
		virtual	void			OnPrevious()			{ if( !IsReadOnly() ) *mSetting = !*mSetting; }

		virtual const char *	GetSettingName() const	{ return (*mSetting) ? "Enabled" : "Disabled"; }

	private:
		bool *					mSetting;
	};
}

//*************************************************************************************
//
//*************************************************************************************
class IRomSettingsComponent : public CRomSettingsComponent
{
	public:

		IRomSettingsComponent( CUIContext * p_context, CFunctor1< const RomSettings & > * on_settings_confirmed );
		~IRomSettingsComponent();

		// CPspRomSelettings
		virtual void				SetSettings( const RomSettings * p_settings );
		virtual void				GetSettings( RomSettings * p_settings );

		// CUIComponent
		virtual void				Update( float elapsed_time, const v2 & stick, u32 old_buttons, u32 new_buttons );
		virtual void				Render();

	private:
		CFunctor1< const RomSettings & > *		OnSettingsConfirmed;

		CUIElementBag				mElements;
		u32							mSelectedIdx;

		RomSettings					mRomSettings;
		u32							mCurrentController;
};


//*************************************************************************************
//
//*************************************************************************************
CRomSettingsComponent::CRomSettingsComponent( CUIContext * p_context )
:	CUIComponent( p_context )
{
}

//*************************************************************************************
//
//*************************************************************************************
CRomSettingsComponent::~CRomSettingsComponent()
{
}

//*************************************************************************************
//
//*************************************************************************************
CRomSettingsComponent *	CRomSettingsComponent::Create( CUIContext * p_context, CFunctor1< const RomSettings & > * on_settings_confirmed )
{
	return new IRomSettingsComponent( p_context, on_settings_confirmed );
}

//*************************************************************************************
//
//*************************************************************************************
IRomSettingsComponent::IRomSettingsComponent( CUIContext * p_context, CFunctor1< const RomSettings & > * on_settings_confirmed )
:	CRomSettingsComponent( p_context )
,	OnSettingsConfirmed( on_settings_confirmed )
,	mSelectedIdx( 0 )
,	mCurrentController( 0 )
{
	mElements.Add( new CSaveTypeSetting( &mRomSettings.SaveType, "Save Type", "The save game mechanism used for this rom (this setting is read-only)" ) );
	mElements.Add( new CBoolSetting( &mRomSettings.ExpansionPakEnabled, "Expansion Pak", "Whether the expansion pak used or required by this rom (not required for many roms)", "Yes", "No" ) );
	mElements.Add( new CBoolSetting( &mRomSettings.DynarecEnabled, "Dynamic Recompilation", "Whether dynamic recompilation is enabled for this rom", "Enabled", "Disabled" ) );
	mElements.Add( new CTextureHashFrequency( &mRomSettings.CheckTextureHashFrequency, "Texture Update Check",	"Whether to check for texture updates between frames. Disable this to improve framerate, at the expense of graphics quality." ) );
	mElements.Add( new CAdjustControllerSetting( &mCurrentController, "Controller" ) );

	mElements.Add( new CBoolSetting( &gAudioPluginEnabled, "Audio", "Whether or not to enable audio emulation.", "Enabled", "Disabled" ) );
	mElements.Add( new CAdjustFrequencySetting( &gAdaptFrequency, " Adjust Frequency", "When enabled, this mode tries to avoid gaps in the audio by adjusting the pitch of the audio stream." ) );
}

//*************************************************************************************
//
//*************************************************************************************
IRomSettingsComponent::~IRomSettingsComponent()
{
	delete OnSettingsConfirmed;
}

//*************************************************************************************
//
//*************************************************************************************
void	IRomSettingsComponent::Update( float elapsed_time, const v2 & stick, u32 old_buttons, u32 new_buttons )
{
	if(old_buttons != new_buttons)
	{
		if(new_buttons & PSP_CTRL_START)
		{
			// This is a bit nasty. We need to return this as a part of RomSettings,
			// but I want to make a distinction between 'absolute' information about
			// a rom (i.e. the save type) and user-specified preferences.
			CInputManager::Get()->SetConfiguration( mCurrentController );

			if(OnSettingsConfirmed != NULL)
			{
				(*OnSettingsConfirmed)( mRomSettings );
			}
		}

		if( new_buttons & PSP_CTRL_UP )
		{
			if( mSelectedIdx > 0 )
			{
				mSelectedIdx--;
			}
		}
		if( new_buttons & PSP_CTRL_DOWN )
		{
			if( mSelectedIdx < mElements.GetNumElements() - 1 )
			{
				mSelectedIdx++;
			}
		}

		if( mSelectedIdx < mElements.GetNumElements() )
		{
			CUIElement *	element( mElements.GetElement( mSelectedIdx ) );
			if( new_buttons & PSP_CTRL_LEFT )
			{
				element->OnPrevious();
			}
			if( new_buttons & PSP_CTRL_RIGHT )
			{
				element->OnNext();
			}
			if( new_buttons & (PSP_CTRL_CROSS/*|PSP_CTRL_START*/) )
			{
				element->OnSelected();
			}
		}
	}
}

//*************************************************************************************
//
//*************************************************************************************
void	IRomSettingsComponent::Render()
{
	mElements.Draw( mpContext, TEXT_AREA_LEFT, TEXT_AREA_RIGHT, TEXT_AREA_TOP, mSelectedIdx );

	if( mSelectedIdx < mElements.GetNumElements() )
	{
		const char *		p_description( mElements.GetElement( mSelectedIdx )->GetDescription() );

		mpContext->DrawTextArea( DESCRIPTION_AREA_LEFT,
								 DESCRIPTION_AREA_TOP,
								 DESCRIPTION_AREA_RIGHT - DESCRIPTION_AREA_LEFT,
								 DESCRIPTION_AREA_BOTTOM - DESCRIPTION_AREA_TOP,
								 p_description,
								 DrawTextUtilities::TextWhite,
								 CUIContext::VA_BOTTOM );
	}
}

//*************************************************************************************
//
//*************************************************************************************
void	IRomSettingsComponent::SetSettings( const RomSettings * p_settings )
{
	mRomSettings = *p_settings;
}

//*************************************************************************************
//
//*************************************************************************************
void	IRomSettingsComponent::GetSettings( RomSettings * p_settings )
{
	*p_settings = mRomSettings;
}

